"""This module contains a ``Configurator`` object.
    The configurator is an object that stores *parameters* that can be
    relevant to make the evolution of a simulation. The objective is to
    introduce flexibility when configuring the the simulator platform
    and the runtime execution possibilities offered by exposed functions
    from the c library of SymuVia
"""
# ============================================================================
# STANDARD  IMPORTS
# ============================================================================

from ctypes import create_string_buffer, c_bool, c_char
from dataclasses import dataclass
import click

# ============================================================================
# INTERNAL IMPORTS
# ============================================================================

from symupy.utils.constants import (
    BUFFER_STRING,
    WRITE_XML,
    TRACE_FLOW,
    DEFAULT_PATH_SYMUVIA,
    TOTAL_SIMULATION_STEPS,
    LAUNCH_MODE,
)

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================


@dataclass
class Configurator:
    """Configurator class for containing specific simulator parameters
        Example:
            To use the ``Simulator`` declare in a string the ``path`` to the
            simulator ::

                >>> path = "path/to/libSymuyVia.dylib"
                >>> simulator = Configurator(library_path = path)

        Args:
            library_path (str):
                Absolute path towards the simulator library

            bufferSize (int):
                Size of the buffer for message for data received from simulator

            write_xml (bool):
                Flag to turn on writting the XML output

            trace_flow (bool):
                Flag to determine tracing or not the flow / trajectories

            total_steps (int):
                Define the number of iterations of a simulation

            step_launch_mode (str):
                Determine to way to launch the ``RunStepEx``. Options ``lite``/``full``

        :return: Configurator object with simulation parameters
        :rtype: Configurator
    """

    buffer_string: c_char = create_string_buffer(BUFFER_STRING)
    write_xml: c_bool = c_bool(WRITE_XML)
    trace_flow: bool = TRACE_FLOW
    library_path: str = DEFAULT_PATH_SYMUVIA
    total_steps: int = TOTAL_SIMULATION_STEPS
    step_launch_mode: str = LAUNCH_MODE

    def __init__(self, **kwargs) -> None:
        """ Configurator class for containing specific simulator parameter

            Args:

            buffer_string (int):
                Size of the buffer for message for data received from simulator

            write_xml (bool):
                Flag to turn on writting the XML output

            trace_flow (bool):
                Flag to determine tracing or not the flow / trajectories

            library_path (str):
                Absolute path towards the simulator library

            total_steps (int):
                Define the number of iterations of a simulation

            step_launch_mode (str):
                Determine to way to launch the ``RunStepEx``. Options ``lite``/``full``
        """
        click.echo("Configurator: Initialization")
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        data_dct = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"{self.__class__.__name__}({data_dct})"

    def __str__(self):
        data_dct = "Configuration status:\n " + "\n ".join(
            f"{k}:  {v}" for k, v in self.__dict__.items()
        )
        return f"{data_dct}"



"""
**Scenario Module**

    This module contains descriptions that stablish a traffic scenario. A traffic scenario 
    is regularly described by a simulation object that points towards properties of the simulator. 
"""

# ============================================================================
# STANDARD  IMPORTS
# ============================================================================

import os
from lxml import etree
from datetime import datetime

# ============================================================================
# INTERNAL IMPORTS
# ============================================================================

from symupy.utils.exceptions import SymupyFileLoadError
from symupy.utils import constants as ct

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================


class Simulation(object):
    def __init__(self, file_name: str) -> None:
        if os.path.exists(file_name):
            self._file_name = file_name
            self.load_xml_tree()
        else:
            raise SymupyFileLoadError("File not found", file_name)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.filename()})"

    def load_xml_tree(self) -> None:
        """ Load XML file_name
        """
        # TODO: Add validation with DTD
        tree = etree.parse(self._file_name)
        root = tree.getroot()
        self._xml_tree = root

    @property
    def xmltree(self):
        return self._xml_tree

    @xmltree.setter
    def xmltree(self, rootxml):
        self._xml_tree = rootxml

    def get_simulation_parameters(self) -> tuple:
        """ Get simulation parameters

        :return: tuple with XML dictionary containing parameters
        :rtype: tuple
        """
        branch_tree = "SIMULATIONS"
        sim_params = self.xmltree.xpath(branch_tree)[0].getchildren()
        return tuple(par.attrib for par in sim_params)

    def get_vehicletype_information(self) -> tuple:
        """ Get the vehicle parameters

        :return: tuple of dictionaries containing vehicle parameters
        :rtype: tuple
        """
        branch_tree = "TRAFICS/TRAFIC/TYPES_DE_VEHICULE"
        vehicle_types = self.xmltree.xpath(branch_tree)[0].getchildren()
        return tuple(v.attrib for v in vehicle_types)

    def get_network_endpoints(self) -> tuple:
        """ Get networks endpoint names

        :return: tuple containing endpoint names
        :rtype: tuple
        """
        branch_tree = "TRAFICS/TRAFIC/EXTREMITES"
        end_points = self.xmltree.xpath(branch_tree)[0].getchildren()
        return tuple(ep.attrib["id"] for ep in end_points)

    def get_network_links(self) -> tuple:
        """ Get network link names

        :return: tuple containing link names 
        :rtype: tuple
        """
        branch_tree = "TRAFICS/TRAFIC/TRONCONS"
        links = self.xmltree.xpath(branch_tree)[0].getchildren()
        return tuple(ep.attrib["id"] for ep in links)

    def get_simulation_steps(self, simid: int = 0) -> range:
        """Get simulation steps for an simulation. specify the simulation id  via an integer value

        :param simid: simulation id , defaults to 0
        :type simid: int, optional
        :return:
        :rtype: range
        """
        t1 = datetime.strptime(self.get_simulation_parameters()[simid].get("debut"), ct.HOUR_FORMAT)
        t2 = datetime.strptime(self.get_simulation_parameters()[simid].get("fin"), ct.HOUR_FORMAT)
        t = t2 - t1
        n = t.seconds / float(self.get_simulation_parameters()[simid].get("pasdetemps"))
        return range(int(n))

    def get_mfd_sensor_names(self) -> tuple:
        """ Get MFD sensors defined for a specific simulation
        
        :return: tuple of MFD sensors in the network
        :rtype: tuple
        """
        branch_tree = "TRAFICS/TRAFIC/PARAMETRAGE_CAPTEURS/CAPTEURS"
        sensors = self.xmltree.xpath(branch_tree)[0].getchildren()
        return tuple(sn.attrib["id"] for sn in sensors)

    def get_links_in_mfd_sensor(self, sensor_id: str) -> tuple:
        """ Get links associated to a particular MFD sensor for a specific simulation
        
        :param sensor_id: Sensor id
        :type sensor_id: str
        :return: tuple of strings with links covered by the sensor
        :rtype: tuple
        """
        branch_tree = "TRAFICS/TRAFIC/PARAMETRAGE_CAPTEURS/CAPTEURS"
        sensors = self.xmltree.xpath(branch_tree)[0].getchildren()
        try:
            sensor_element = sensors[self.get_mfd_sensor_names().index(sensor_id)]
            links = sensor_element.getchildren()[0].getchildren()
            return tuple(lk.attrib["id"] for lk in links)
        except:
            return ()

    def __contains__(self, value: tuple) -> bool:
        # REVIEW: Implement? in method? maybe useful
        raise NotImplementedError

    def filename(self, encoding: str = None):
        """ 
            This method returns the value of encoding of the simulation scenario under consideration
        
            :param encoding: enconder UTF8, defaults to None
            :type encoding: string, optional
            :return: Full path of scenario
            :rtype: string
        """
        if encoding == "UTF8":
            return self._file_name.encode(encoding)
        return self._file_name

    @property
    def time_step(self):
        return float(self.get_simulation_parameters()[0].get("pasdetemps"))



import typing
import numpy as np
from symupy.utils import constants as CT

PAR = {"time_step": CT.DCT_SIMULATION_INFO, "engine_tau": CT.ENGINE_CONSTANT}


def dynamic_3rd_ego(state: np.array, control: np.array, parameters=PAR) -> np.array:
    """Update vehicle state in 3rd order dynamics"""
    K_a = parameters["time_step"] / parameters["engine_tau"]
    A = np.array([[1, parameters["time_step"], 0], [0, 1, parameters["time_step"]], [0, 0, (1 - K_a)],])
    B = np.array([[0], [0], [K_a]])
    return A @ state[:3] + B @ control[:1]


def dynamic_2nd_ego(state: np.array, control: np.array, parameters=PAR) -> np.array:
    """Update vehicle state in 2nd order dynamics"""
    A = np.array([[1, parameters["time_step"]], [0, 1]])
    B = np.array([[0], [parameters["time_Step"]]])
    return A @ state[:2] + B @ control[:1]


class VehicleDynamic(object):
    def __init__(self, time_step=CT.TIME_STEP, veh_dyn=dynamic_2nd_ego) -> None:
        self.time_step = time_step
        self.veh_dyn = veh_dyn
        self.prev_state = np.array([])

    def __call__(self, *args, **kwargs):

        vehicle, control, parameters = args
        DCT_DIRECTIVES = {
            "dynamic_3rd_ego": {"args": (vehicle.state, control, parameters), "kwargs": kwargs,},
            "dynamic_2nd_ego": {"args": (vehicle.state, control), "kwargs": kwargs},
        }

        dynamic_name = self.veh_dyn.__name__
        args = DCT_DIRECTIVES.get(dynamic_name).get("args")
        kwargs = DCT_DIRECTIVES.get(dynamic_name).get("kwargs")
        return self.veh_dyn(*args, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}(time_step = {self.time_step}, veh_dyn ={self.veh_dyn.__name__})"

    def __str__(self):
        return f"{self.__class__.__name__}(time_step = {self.time_step}, veh_dyn ={self.veh_dyn.__name__})"



""" 
    This module details the implementation of a ``Simulator`` object in charge of handling the connection between the traffic simulator and this interface. The connection with the traffic simulator is handled by an object called ``Connector`` which establishes a messaging protocol with the traffic simulator. 

    Example:
        To use the ``Simulator`` declare in a string the ``path`` to the simulator ::

            >>> from symupy.api import Simulator
            >>> path_symuvia = "path/to/libSymuyVia.dylib"
            >>> simulator = Simulator(library_path=path_symuvia)

    Other parameters can also be send to the simulator in order to provide other configurations:

    Example: 
        To send make increase the *buffer size* to a specific size:
            
            >>> simulator = Simulator(bufferSize = 1000000)
        
        To increase change the flag that traces the flow:

            >>> simulator = Simulator(trace_flow = True)
"""

# ============================================================================
# STANDARD  IMPORTS
# ============================================================================

import os
from itertools import repeat
from ctypes import cdll, create_string_buffer, c_int, byref, c_bool, c_double
import click
import platform

import typing
from typing import Union

# ============================================================================
# INTERNAL IMPORTS
# ============================================================================

# Error Handling
from symupy.utils.exceptions import (
    SymupyLoadLibraryError,
    SymupyFileLoadError,
    SymupyVehicleCreationError,
    SymupyDriveVehicleError,
    SymupyWarning,
)

#
from .scenario import Simulation
from symupy.utils import SimulatorRequest, Configurator
from symupy.logic import RuntimeDevice

from symupy.utils import timer_func, printer_time
from symupy.utils import constants as CT

from symupy.components import V2INetwork, V2VNetwork
from symupy.components import Vehicle

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================

# V2X Connectivity
NetworkType = Union[V2INetwork, V2VNetwork]


class Simulator(Configurator, RuntimeDevice):
    """ 
        Simulator class for containing object to connect and  command a simulation in SymuVia

        Example:
            Call of the default simulator ::

                >>> from symupy.api import Simulator
                >>> simulator = Simulator()

        :return: Symuvia simulator object with simulation parameters
        :rtype: Simulator

        You may also pass suplementary parameters to the object by specifying keys in the call: 

        Example: 
            To use the ``Simulator`` declare in a string the ``path`` to the simulator ::

                >>> from symupy.api import Simulator
                >>> path_symuvia = "path/to/libSymuyVia.dylib"
                >>> simulator = Simulator(library_path=path_symuvia)
    
        This object describes is a configurator manager for the interface between the traffic simulator and the python interface. For more details on the optinal keyword parameters please refer to :py:class:`~symupy.utils.configurator.Configurator` class.

        :raises SymupyLoadLibraryError: 
            Error raised whenever the SymuVia library is not found

        :raises SymupyFileLoadError: 
            Error raised whenever the provided path for an scenario cannot be loaded into the Simulator

        :raises SymupyVehicleCreationError: 
            Error raised when a vehicle cannot be created

        :raises SymupyDriveVehicleError: 
            Error rased when a vehicle state cannot be imposed

        :raises NotImplementedError: 
            Not implemented functionality 

        :return: Simulator manager object 

        :rtype: Simulator
    """

    def __init__(self, **kwargs) -> None:
        Configurator.__init__(self, **kwargs)
        RuntimeDevice.__init__(self)
        self._net = []

    def __repr__(self):
        return f"{self.__class__.__name__}({self.library_path})"

    # ============================================================================
    # LOADING METHODS
    # ============================================================================

    def load_symuvia(self) -> None:
        """ load SymuVia shared library """
        try:
            lib_symuvia = cdll.LoadLibrary(self.library_path)
        except OSError:
            raise SymupyLoadLibraryError("Library not found", self.library_path)
        self.__library = lib_symuvia

    def load_network(self) -> int:
        """ load SymuVia Simulation File """
        if not hasattr(self, "_sim"):
            raise SymupyFileLoadError("File not provided", "")
        valid = self.__library.SymLoadNetworkEx(self.scenarioFilename("UTF8"))
        if not valid:
            raise SymupyFileLoadError("Simulation could not be loaded", "")
        return valid

    def register_simulation(self, scenarioPath: str) -> None:
        """
            Register simulation file within the simulator

            :param scenarioPath: Path to scenario 
            :type scenarioPath: str
        """
        self._sim = Simulation(scenarioPath)

    def register_network(self, network: NetworkType) -> None:
        # TODO: Impleement this connection. This is for V2V
        self._net.append(network)

    # ============================================================================
    # RUNTIME METHODS
    # ============================================================================

    @timer_func
    def run_simulation(self, sim_object: Simulation = "") -> None:
        """ Run simulation in a single shot

            Args:
                sim_object (Simulation): Valid simulation scenario

            Returns:
                None: No returns provided, only internal updates
        """
        if sim_object:
            self.register_simulation(sim_object)

        self.load_symuvia()
        self.__library.SymRunEx(self.scenarioFilename("UTF8"))

    def run(self, sim_object: Simulation = "") -> None:
        """ Alias metho to run simulation
            Args:
                sim_object (Simulation): Valid simulation scenario

            Returns:
                None: No returns provided, only internal updates
        """
        self.run_simulation(sim_object)

    def request_answer(self):
        """Request simulator answer and maps the data locally
        """
        if self.step_launch_mode == "lite":
            self._bContinue = self.__library.SymRunNextStepLiteEx(
                self.write_xml, byref(self._b_end)
            )
            return
        self._bContinue = self.__library.SymRunNextStepEx(
            self.buffer_string, self.write_xml, byref(self._b_end)
        )
        self.request.parse_data(self.buffer_string)

    # @printer_time
    def run_step(self) -> int:
        """ Run simulation step by step

        :return: iteration step
        :rtype: int
        """
        try:
            self.request_answer()
            self._c_iter = next(self._n_iter)
            return self._c_iter
        except StopIteration:
            self._bContinue = False
            return -1

    def stop_step(self):
        """Stop current current step of running simulation
        """
        self._bContinue = False

    def create_vehicle(
        self,
        vehtype: str,
        origin: str,
        destination: str,
        lane: int = 1,
        simid: int = 0,
    ) -> int:
        """Creates a vehicle within the network

        :param vehtype: vehicle type according to simulation definitions
        :type vehtype: str
        :param origin: network endpoint nodeaccording to simulation
        :type origin: str
        :param destination: network endpoint nodeaccording to simulation
        :type destination: str
        :param lane: vehicle lane number, defaults to 1
        :type lane: int, optional
        :param simid: simulation id, defaults to 0
        :type simid: int, optional
        :raises SymupyVehicleCreationError: Exception handling for invalid vehicle types or invalid network points
        :return: Vehicle id of the vehicle created >0
        :rtype: int
        """
        endpoints = self._sim.get_network_endpoints()
        veh_data = self._sim.get_vehicletype_information()
        dbTime = self.simulationstep
        vehid = tuple(v["id"] for v in veh_data)

        # Consistency checks
        if vehtype not in vehid:
            raise SymupyVehicleCreationError(
                "Unexisting Vehicle Class in File: ", self.scenarioFilename()
            )

        if (origin not in endpoints) or (destination not in endpoints):
            raise SymupyVehicleCreationError(
                "Unexisting Network Endpoint File: ", self.scenarioFilename()
            )

        # Vehicle creation
        vehid = self.__library.SymCreateVehicleEx(
            vehtype.encode("UTF8"),
            origin.encode("UTF8"),
            destination.encode("UTF8"),
            c_int(lane),
            c_double(dbTime),
        )
        return vehid

    def create_vehicle_with_route(
        self,
        vehtype: str,
        origin: str,
        destination: str,
        lane: int = 1,
        creation_time: float = 0,
        route: str = "",
    ) -> int:
        """ Creates a vehicle with a specific route
        
            :param vehtype: vehicle type according to simulation definitions
            :type vehtype: str
            :param origin: network endpoint node according to simulation
            :type origin: str
            :param destination: network endpoint node according to simulation
            :type destination: str
            :param lane: vehicle lane number, defaults to 1
            :type lane: int, optional
            :param creation_time: time instant of creation [0,Ts], defaults to 0
            :type creation_time: float, optional
            :param route: route followed by the vehicle, defaults to ""
            :type route: str, optional

            :return: Vehicle id of the vehicle created >0
            :rtype: int        
        """
        if origin == destination:
            return -1

        endpoints = self._sim.get_network_endpoints()
        veh_data = self._sim.get_vehicletype_information()
        dbTime = self.simulationstep
        vehid = tuple(v["id"] for v in veh_data)

        # Consistency checks
        if vehtype not in vehid:
            raise SymupyVehicleCreationError(
                "Unexisting Vehicle Class in File: ", self.scenarioFilename()
            )

        if (origin not in endpoints) or (destination not in endpoints):
            raise SymupyVehicleCreationError(
                "Unexisting Network Endpoint File: ", self.scenarioFilename()
            )

        # Vehicle creation
        vehid = self.__library.SymCreateVehicleWithRouteEx(
            origin.encode("UTF8"),
            destination.encode("UTF8"),
            vehtype.encode("UTF8"),
            c_int(lane),
            c_double(creation_time - dbTime),
            route.encode("UTF8"),
        )
        return vehid

    def drive_vehicle(
        self, vehid: int, new_pos: float, destination: str = None, lane: str = 1
    ) -> None:
        """Drives a vehicle to a specific position

        :param vehid: vehicle id to drive 
        :type vehid: int
        :param new_pos: position to place the vehicle
        :type new_pos: float
        :param destination: link of destination, defaults to None
        :type destination: str, optional
        :param lane: lane fo destination, defaults to 1
        :type lane: str, optional
        :raises SymupyDriveVehicleError: Raises error when link does not exist
        :return: Value returned by SymDriveErr 1 ok, negative values are errors.
        :rtype: None
        """
        links = self._sim.get_network_links()

        if not destination:
            destination = self.request.query_vehicle_link(str(vehid))[0]

        if destination not in links:
            raise SymupyDriveVehicleError(
                "Unexisting Network Endpoint File: ", self.scenarioFilename()
            )

        # TODO: Validate that position do not overpass the max pos
        dr_state = self.__library.SymDriveVehicleEx(
            c_int(vehid),
            destination.encode("UTF8"),
            c_int(lane),
            c_double(new_pos),
            1,
        )
        self.request_answer()
        return dr_state

    def drive_vehicle_with_control(
        self, vehcontrol, vehid: int, destination: str = None, lane: str = 1
    ):
        # TODO: Basic prototyping
        vehcontrol.set_current_state(self.request)
        new_pos = vehcontrol.new_position
        return self.drive_vehicle(vehid, new_pos, destination, lane)

    def get_vehicle_context(self, vehid: str):
        ## TODO: Implement this
        raise NotImplementedError

    def log_vehicle_in_network(self, veh: Vehicle, network: NetworkType):
        # veh = Vehicle(vehid)
        ## TODO: Finish
        network.register_vehicle(veh)

    def log_vehid_in_network(self, vehid: str, network: NetworkType):
        ## TODO: Optional
        pass

    def init_total_travel_time(self):
        """ Counter initializer for total travel time
        """
        # TODO: Improvement → Better organizadtion
        self.__library.SymGetTotalTravelTimeEx.restype = c_double

    def init_total_travel_distance(self):
        """ Counter initializer for total travel time
        """
        # TODO: Improvement → Better organizadtion
        self.__library.SymGetTotalTravelDistanceEx.restype = c_double

    def get_total_travel_time(self, sensors_mfd: list = []):
        """ Computes the total travel time of vehicles in a MFD region
        
        :param zone_id: MFD sensor id, defaults to None
        :type zone_id: str, optional
        :return: Associated total travel time
        """
        # TODO: Improvement → Better organizadtion
        if isinstance(sensors_mfd, str):
            return self.__library.SymGetTotalTravelTimeEx(
                sensors_mfd.encode("UTF8")
            )

        if not sensors_mfd:
            sensors_mfd = self.simulation.get_mfd_sensor_names()

        return tuple(
            self.__library.SymGetTotalTravelTimeEx(sensor.encode("UTF8"))
            for sensor in sensors_mfd
        )

    def get_total_travel_distance(self, sensors_mfd: list = []):
        """ Computes the total travel distance of vehicles in a MFD region
        
        :param sensors_mfd: MFD sensor ids, defaults to None
        :type sensors_mfd: list, optional
        :return: Associated total travel distance
        """
        # TODO: Improvement → Better organizadtion
        if isinstance(sensors_mfd, str):
            return self.__library.SymGetTotalTravelDistanceEx(
                sensors_mfd.encode("UTF8")
            )

        if not sensors_mfd:
            sensors_mfd = self.simulation.get_mfd_sensor_names()

        return tuple(
            self.__library.SymGetTotalTravelDistanceEx(sensor.encode("UTF8"))
            for sensor in sensors_mfd
        )

    def get_mfd_speed(self, sensors_mfd: list = []):
        """ Computes the total speed of vehicles in a MFD region
        
        :param sensors_mfd: MFD sensor ids, defaults to None
        :type sensors_mfd: list, optional
        :return: speed computed as ttt/ttd
        """
        # TODO: Improvement → Better organizadtion
        if isinstance(sensors_mfd, str):
            d = self.get_total_travel_distance(sensors_mfd)
            t = self.get_total_travel_time(sensors_mfd)
            spd = d / t if t != 0 else 10
            return spd

        itdsttm = zip(
            self.get_total_travel_distance(sensors_mfd),
            self.get_total_travel_time(sensors_mfd),
        )
        spd = []
        for d, t in itdsttm:
            if t != 0:
                spd.append(d / t)
            else:
                spd.append(10)  # minimum speed?
        return tuple(spd)

    def add_control_probability_zone_mfd(
        self, access_probability: dict, minimum_distance: dict
    ) -> None:
        """
            Add a probability to control the access to a specific zone within the network
        
            :param access_probability: Key (zone name) Value (probability of access)
            :type access_probability: dict
            :param minimum_distance: Key (zone name) Value (distance before entering the zone to activate policy)
            :type minimum_distance: dict
        """
        self.dctidzone = {}

        for tp_zn_pb, tp_zn_md in zip(
            access_probability.items(), minimum_distance.items()
        ):
            sensor, accrate = tp_zn_pb
            _, min_dst = tp_zn_md
            links = self.simulation.get_links_in_mfd_sensor(sensor)
            links_str = " ".join(links)
            self.dctidzone[sensor] = self.__library.SymAddControlZoneEx(
                -1,
                c_double(accrate),
                c_double(min_dst),
                f"{links_str}".encode("UTF8"),
            )
        # Apply set control
        self.__library.SymApplyControlZonesEx(-1)
        return self.dctidzone

    def modify_control_probability_zone_mfd(
        self, access_probability: dict
    ) -> None:
        """
            Modifies a probability to control the access to a specific zone within the network
        
            :param access_probability: Key (zone name) Value (probability of access)
            :type access_probability: dict
        """

        for sensor, probablity in access_probability.items():
            self.__library.SymModifyControlZoneEx(
                -1, self.dctidzone[sensor], c_double(probablity)
            )
        # Apply set control
        self.__library.SymApplyControlZonesEx(-1)
        return self.dctidzone

    def __enter__(self) -> None:
        """
            This method initializes the usage of the ``Simulator`` class as a context manager. 

            The protocol followed in order to perform the full connection is as follows 

        """

        self.reset_state()

        # Compliance situation
        self.__performCompliance()

        # Connect to platform
        self.__performConnect()

        # Variable initialization
        self.__performInitialize()

        # Extra

        return self

    def __exit__(self, type, value, traceback) -> bool:
        self.__library.SymUnloadCurrentNetworkEx()
        click.echo("Runtime: End")
        return False

    def build_dynamic_param(self):
        """Construct parameters for vehicle dynamics
        """
        self.__dct_par = {
            "time_step": self.simulation.time_step,
            "engine_tau": CT.ENGINE_CONSTANT,
        }

    # ============================================================================
    # STATE MACHINE
    # ============================================================================

    def __performCompliance(self) -> None:
        """
            Perform compliance check
        """
        self.next_state(True)

    def __performConnect(self) -> None:
        """
            Perform simulation connection
        """
        self.load_symuvia()
        self.load_network()
        self.next_state(True)

    def __performInitialize(self) -> None:
        """
            Perform simulation initialization
        """
        self._b_end = c_int()
        self.request = SimulatorRequest()
        self._n_iter = iter(self._sim.get_simulation_steps())
        self._c_iter = next(self._n_iter)
        self._bContinue = True

        self.init_total_travel_distance()
        self.init_total_travel_time()
        self.build_dynamic_param()

        self.next_state(self.do_next)

    def __performPreRoutine(self) -> None:
        """
            Perform simulator preroutine
        """
        self.next_state(self.do_next)

    def __performQuery(self) -> None:
        """
            Perform simulator Query
        """
        self.next_state(self.do_next)

    def __performControl(self) -> None:
        """
            Perform simulator Control
        """
        self.next_state(self.do_next)

    def _set_manual_initialization(self) -> None:
        """
            This method is a way to set manual initialization of the simulator
            for testing purposes. (Internal use)
        """
        self.__performCompliance()
        self.__performConnect()
        self.__performInitialize()

    # ============================================================================
    # ATTRIBUTES
    # ============================================================================

    def scenarioFilename(self, encoding=None) -> str:
        """ 
            Scenario filenamme
        
            :return: Absolute path towards the XML input for SymuVia
            :rtype: str
        """
        return self.simulation.filename(encoding)

    @property
    def s_response_dec(self):
        """ 
            Obtains instantaneous data from simulator

            :return: last query from simulator
            :rtype: str
        """
        return self.buffer_string.value.decode("UTF8")

    @property
    def do_next(self) -> bool:
        """
            Returns true if the simulation shold continue
        
            :return: True if next step continues
            :rtype: bool
        """
        return self._bContinue

    @property
    def get_request(self) -> dict:
        """
            Returns the query received from the simulator
        
            :return: Request from the simulator
            :rtype: dict
        """
        return self.request.data_query

    @property
    def simulation(self) -> Simulation:
        """
            Simulation scenario 
        
            :return: Object describing senario under simulation
            :rtype: Simulation
        """
        return self._sim

    @property
    def simulationstep(self) -> str:
        """ 
            Current simulation step.

            Example:
                You can use the time step to control actions 

                >>> with simulator as s:
                ...     while s.do_next()
                ...         if s.simulationstep>0:
                ...             print(s.simulationtimestep)
        
            :return: current simulation iteration
            :rtype: str
        """
        return self._c_iter

    @property
    def sampling_time(self) -> float:
        """ 
            Simulation sampling time 
        
            :return: sampling time from XML file
            :rtype: float
        """
        return self.simulation.sampling_time

    # ============================================================================
    # CONSTRUCTORS
    # ============================================================================

    @classmethod
    def from_path(cls, filename_path, symuvia_path):
        """ Alternative constructor for the Simulator 

            Example:
                To use this alternative constructor ``Simulator`` declare in a string the ``path`` to the simulator ::

                    >>> path = "path/to/simulator.so"
                    >>> scenario = "path/to/scenario.xml"
                    >>> simulator = Simulator.from_path(path,scenario) 

        """
        sim = cls(library_path=symuvia_path)
        sim.register_simulation(filename_path)
        return sim



""" 
    This module handles the Simulation response converting it into proper formats for querying data afterwards. 
"""

# ============================================================================
# STANDARD  IMPORTS
# ============================================================================

from xmltodict import parse
from xml.parsers.expat import ExpatError

# ============================================================================
# INTERNAL IMPORTS
# ============================================================================

from symupy.components import Vehicle, VehicleList

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================


class SimulatorRequest:
    def __init__(self):
        self._str_response = ""
        self._vehs = []

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __str__(self):
        return (
            "Sim Time: {}, VehInNetwork: {}".format(self.current_time, self.current_nbveh)
            if self.data_query
            else "Simulation has not started"
        )

    def parse_data(self, response: str = None) -> dict:
        """Parses response from simulator to data

        :param response: Simulator response
        :type response: str
        :return: Full simulator response
        :rtype: dict
        """
        self._str_response = response

    def get_vehicle_data(self) -> list:
        """Extracts vehicles information from simulators response

        :param response: Simulator response
        :type response: str
        :return: list of vehicles in the network
        :rtype: list of dictionaries
        """
        if self.data_query.get("INST", {}).get("TRAJS") is not None:
            veh_data = self.data_query.get("INST").get("TRAJS")
            if isinstance(veh_data["TRAJ"], list):
                return veh_data["TRAJ"]
            return [veh_data["TRAJ"]]
        return []

    def get_vehicle_id(self) -> tuple:
        """Extracts vehicle ids information from simulators response

        :return: tuple containing vehicle ids at current state in all network
        :rtype: list
        """
        return tuple(veh.get("@id") for veh in self.get_vehicle_data())

    def query_vehicle_link(self, vehid: str, *args) -> tuple:
        """ Extracts current vehicle link information from simulators response

        :param vehid: vehicle id multiple arguments accepted
        :type vehid: str
        :return: vehicle link in tuple form
        :rtype: tuple
        """
        vehids = set((vehid, *args)) if args else vehid
        vehid_pos = self.query_vehicle_data_dict("@tron", vehids)
        return tuple(vehid_pos.get(veh) for veh in vehids)

    def query_vehicle_position(self, vehid: str, *args) -> tuple:
        """ Extracts current vehicle distance information from simulators response

        :param vehid: vehicle id multiple arguments accepted
        :type vehid: str
        :return: vehicle distance in link in tuple form
        :rtype: tuple
        """
        vehids = set((vehid, *args)) if args else vehid
        vehid_pos = self.query_vehicle_data_dict("@dst", vehids)
        return tuple(vehid_pos.get(veh) for veh in vehids)

    def query_vehicle_data_dict(self, dataval: str, vehid: str, *args) -> dict:
        """ Extracts and filters vehicle data from the simulators response

        :param dataval: parameter to be extracted e.g. '@id', '@dst'
        :type dataval: str
        :param vehid: vehicle id, multiple arguments accepted
        :type vehid: str
        :return: dictionary where key is @id and value is dataval
        :rtype: dict
        """
        vehids = set((vehid, *args)) if args else set(vehid)
        data_vehs = [(veh.get("@id"), veh.get(dataval)) for veh in self.get_vehicle_data() if veh.get("@id") in vehids]
        return dict(data_vehs)

    def is_vehicle_in_network(self, vehid: str, *args) -> bool:
        """True if veh id is in the network at current state, for multiple arguments
           True if all veh ids are in the network

        :param vehid: Integer of vehicle id, comma separated if testing for multiple
        :type vehid: int
        :return: True if vehicle is in the network otherwise false
        :rtype: bool
        """
        all_vehs = self.get_vehicle_id()
        if not args:
            return vehid in all_vehs
        vehids = set((vehid, *args))
        return set(vehids).issubset(set(all_vehs))

    def vehicle_in_link(self, link: str, lane: str = "1") -> tuple:
        """Returns a tuple containing vehicle ids traveling on the same link+lane at current state

        :param link: link name
        :type link: str
        :param lane: lane number, defaults to '1'
        :type lane: str, optional
        :return: tuple containing vehicle ids
        :rtype: tuple
        """
        return tuple(
            veh.get("@id") for veh in self.get_vehicle_data() if veh.get("@tron") == link and veh.get("@voie") == lane
        )

    def is_vehicle_in_link(self, veh: str, link: str) -> bool:
        """ Returns true if a vehicle is in a link at current state
        
        :param veh: vehicle id
        :type veh: str
        :param link: link name
        :type link: str
        :return: True if veh is in link
        :rtype: bool
        """
        veh_ids = self.vehicle_in_link(link)
        return set(veh).issubset(set(veh_ids))

    def is_vehicle_driven(self, vehid: str) -> bool:
        """ Returns true if the vehicle state is exposed to a driven state

            Args:
                vehid (str):
                    vehicle id
        """
        if self.is_vehicle_in_network(vehid):

            forced = tuple(
                veh.get("@etat_pilotage") == "force (ecoulement respecte)"
                for veh in self.get_vehicle_data()
                if veh.get("@id") == vehid
            )
            return any(forced)
        return False

    def vehicle_downstream_of(self, vehid: str) -> tuple:
        """Get ids of vehicles downstream to vehid

        :param vehid: integer describing id of reference veh
        :type vehid: int
        :return: tuple with ids of vehicles ahead (downstream)
        :rtype: tuple
        """
        link = self.query_vehicle_link(vehid)[0]
        vehpos = self.query_vehicle_position(vehid)[0]

        vehids = set(self.vehicle_in_link(link))
        neigh = vehids.difference(set(vehid))

        neighpos = self.query_vehicle_position(*neigh)

        return tuple(nbh for nbh, npos in zip(neigh, neighpos) if float(npos) > float(vehpos))

    def vehicle_upstream_of(self, vehid: str) -> tuple:
        """Get ids of vehicles upstream to vehid

        :param vehid: integer describing id of reference veh
        :type vehid: int
        :return: tuple with ids of vehicles behind (upstream)
        :rtype: tuple
        """
        link = self.query_vehicle_link(vehid)[0]
        vehpos = self.query_vehicle_position(vehid)[0]

        vehids = set(self.vehicle_in_link(link))
        neigh = vehids.difference(set(vehid))

        neighpos = self.query_vehicle_position(*neigh)

        return tuple(nbh for nbh, npos in zip(neigh, neighpos) if float(npos) < float(vehpos))

    def create_vehicle_list(self):
        """Initialize 
        """
        if not self._vehs:
            self._vehs = VehicleList.from_request(self.get_vehicle_data())
            return

    def update_vehicle_list(self):
        """ Construct and or update vehicle data
        """
        if self._vehs:
            self._vehs.update_list(self.get_vehicle_data())
            return
        self.create_vehicle_list()

    def __contains__(self, elem: Vehicle) -> bool:
        return elem in self._vehs

    @property
    def data_query(self):
        try:
            return parse(self._str_response.value)
        except ExpatError:
            return {}
        except AttributeError:
            return {}

    @property
    def vehicles(self):
        self.update_vehicle_list()
        return self._vehs

    @property
    def current_time(self) -> str:
        return self.data_query.get("INST").get("@val")

    @property
    def current_nbveh(self) -> int:
        return self.data_query.get("INST").get("@nbVeh")



""" This module computes/modifies data for XML creation for the simulator 
"""
from lxml import etree as ET
from .constants import (
    DCT_SIMULATION_INFO,
    DCT_EXPORT_INFO,
    DCT_TRAFIC_INFO,
    DCT_NETWORK_INFO,
    DCT_SCENARIO_INFO,
)


def create_DCT_SIMULATION_INFO(
    title: str,
    simid: str,
    timestep: float,
    time_start: str,
    time_end: str,
    date: str,
    seed: int,
    cfl: str,
    flow: str,
    decel: bool,
) -> dict:
    """Computes dictionary DCT_SIMULATION_INFO header for XML file 
    
    :param title: Simulation title
    :type title: str
    :param simid: Simulation ID 
    :type simid: str
    :param timestep: time step (seconds)
    :type timestep: float
    :param time_start: start time H%:%M:%S
    :type time_start: str
    :param time_end: start time H%:%M:%S
    :type time_end: str
    :param date: simulation date Y%-%m-%d
    :type date: str
    :param seed: random seed
    :type seed: int
    :param cfl: car following law 
    :type cfl: str
    :param flow: flow behaviour
    :type flow: str
    :param decel: deceleration procedure activation
    :type decel: bool
    :return: dictionary with metadata for xml creation
    :rtype: dict
    """
    DCT_SIMULATION_INFO.update({"id": f"{simid}"})
    DCT_SIMULATION_INFO.update({"pasdetemps": f"{str(timestep)}"})
    DCT_SIMULATION_INFO.update({"debut": f"{time_start}"})
    DCT_SIMULATION_INFO.update({"fin": f"{time_end}"})
    DCT_SIMULATION_INFO.update({"loipoursuite": f"{cfl}"})
    DCT_SIMULATION_INFO.update({"comportementflux": f"{flow}"})
    DCT_SIMULATION_INFO.update({"date": f"{date}"})
    DCT_SIMULATION_INFO.update({"titre": f"{title}"})
    DCT_SIMULATION_INFO.update({"proc_deceleration": f"{str(decel).lower()}"})
    DCT_SIMULATION_INFO.update({"seed": f"{str(seed)}"})

    return DCT_SIMULATION_INFO


def create_DCT_NETWORK_INFO(networkid: str) -> dict:
    """Computes dictionary DCT_NETWORK_INFO for XML file 
    
    :param networkid: network identifier
    :type networkid: str
    :return: dict
    :rtype: [type]
    """
    DCT_NETWORK_INFO.update({"id": networkid})
    return DCT_NETWORK_INFO


def create_DCT_TRAFIC_INFO(
    trafid: str, bounded_acc: bool = True, relaxation: float = 4, ghost_lanechange: bool = False,
) -> dict:
    """Computes dictionary DCT_TRAFIC_INFO for XML file 
    
    :param trafid: traffic identifier
    :type trafid: str
    :param bounded_acc: bounds acceleration True/False
    :type bounded_acc: bool
    :param relaxation: relaxation coefficient lane change
    :type relaxation: float
    :param ghost_lanechange: keep lane change ghost True/False
    :type ghost_lanechange: bool
    :return: dictionary with metadata for xml creation
    :rtype: dict
    """
    DCT_TRAFIC_INFO.update({"id": str(trafid)})
    DCT_TRAFIC_INFO.update({"accbornee": str(bounded_acc)})
    DCT_TRAFIC_INFO.update({"coeffrelax": str(relaxation)})
    DCT_TRAFIC_INFO.update({"chgtvoie_ghost": str(ghost_lanechange)})
    return DCT_TRAFIC_INFO


def create_DCT_EXPORT_INFO(trace_rout: bool = False, trajectories: bool = True, csv: bool = True) -> dict:
    """Computes dictionary DCT_SCENARIO_INFO for XML file
    
    :param trace_rout: trace routes True/False
    :type trace_rout: bool
    :param trajectories: export trajectories True/False
    :type trajectories: bool
    :param csv: export trajectories in csv
    :type csv: bool
    :return: dictionary with metadata for xml creation
    :rtype: dict
    """
    DCT_EXPORT_INFO.update({"trace_route": str(trace_rout).lower()})
    DCT_EXPORT_INFO.update({"trajectoires": str(trajectories).lower()})
    DCT_EXPORT_INFO.update({"csv": str(csv).lower()})
    return DCT_EXPORT_INFO


def create_DCT_SCENARIO_INFO(
    scnid: str, dirout: str, prefout: str, dct_sim: dict, dct_traffic: dict, dct_network: dict,
) -> dict:
    """Computes dictionary DCT_SCENARIO_INFO for XML file
    
    :param scnid: Scenario identifier
    :type scnid: str
    :param dirout: Output directory
    :type dirout: str
    :param prefout: Prefix
    :type prefout: str
    :param dct_sim: DCT_SIMULATION_INFO
    :type dct_sim: dict
    :param dct_traffic: DCT_TRAFIC_INFO
    :type dct_traffic: dict
    :param dct_network: DCT_NETWORK_INFO
    :type dct_network: dict
    :return: dictionary with metadata for xml creation
    :rtype: dict
    """
    DCT_SIMULATION_INFO.update({"id": f"{scnid}"})
    DCT_SIMULATION_INFO.update({"simulation_id": f"{dirout}"})
    DCT_SIMULATION_INFO.update({"trafic_id": f"{prefout}"})
    DCT_SIMULATION_INFO.update({"reseau_id": f"{dct_sim.get('id')}"})
    DCT_SIMULATION_INFO.update({"dirout": f"{dct_traffic.get('id')}"})
    DCT_SIMULATION_INFO.update({"prefout": f"{dct_network.get('id')}"})
    return DCT_SCENARIO_INFO


def fix_values(element: ET.ElementTree, dct_values: dict) -> ET.ElementTree:
    """ Fix a set of values within a dictionary as attributes of an XML subelement
    
    :param element: XML tree element
    :type element: ET.ElementTree
    :param dct_values: attributes to fix
    :type dct_values: dict
    :return: Tranformed XML tree element
    :rtype: ET.ElementTree
    """
    for k, v in dct_values.items():
        element.set(k, v)
    return element


class XMLGenerator(object):
    """ A class to handle XML generator for SymuVia
    
    """

    xmlns = "http://www.w3.org/2001/XMLSchema-instance"
    namespace = "noNamespaceSchemaLocation"
    nsmap = {"xsi": xmlns}

    def __init__(self, path_xsd: str = ""):
        # Create a general scheme
        attr_qname = ET.QName(self.xmlns, self.namespace)
        self.root = ET.Element("ROOT_SYMUBRUIT", {attr_qname: "reseau.xsd"}, version="2.05", nsmap=self.nsmap,)



"""
    A class to take control over vehicles 
"""

from dataclasses import dataclass, field
from collections import deque
from typing import List

from symupy.components import Vehicle, VehicleList
from symupy.utils import constants as ct

import numpy as np


class VehicleControl:
    """ Controls a single vehicle"""

    def __init__(self, vehicle: Vehicle, mode: str = "manual", time: float = ct.TIME_STEP):
        self.mode = mode
        self.time = time
        self.vehicle = vehicle
        self.__pos = vehicle.distance
        self.__spd = vehicle.speed
        self.__acc = vehicle.acceleration

    def __repr__(self):
        return f"{self.__class__.__name__}(vehicle= {self.vehicle.__str__()},mode = {self.mode}, time= {self.time})"

    def __str__(self):
        return f"{self.__class__.__name__}(vehicle= {self.vehicle.__str__()},mode = {self.mode}, time= {self.time})"

    def compute_control(self):
        """Compute automatic control"""
        raise NotImplementedError

    def set_manual_control(self, value, controlvar="acceleration"):
        self.control = {"variable": controlvar, "value": value}

    @property
    def control(self):
        return self.__u

    @control.setter
    def control(self, value):
        """ Setter for manual """
        if self.mode == "manual":
            self.__u = value
            return
        self.__u = self.compute_control()


@dataclass
class VehicleGroupControl:
    """ Control of group of vehicles of the same class"""

    controls: List[VehicleControl] = field(default_factory=list)

    def __str__(self):
        if not self.controls:
            return "No vehicles registered for control"
        return "\n".join(", ".join(f"{k}:{v}" for k, v in ctr.vehicles.__dict__.items()) for ctr in self.controls)

    def __repr__(self):
        if not self.controls:
            return "No vehicles registered for control"
        return "\n".join(", ".join(f"{k}:{v}" for k, v in ctr.vehicles.__dict__.items()) for ctr in self.controls)



"""This module contains a set of exceptions and handlers for such exceptions.

    Example:
        To use the ``Exceptions`` import the module as::

            >>> from symupy.utils.exceptions import SymupyLoadLibraryError
            >>> from ctypes import cdll
            >>> try:
            ...     path = 'path/to/library'
            ...     lib = cdll.LoadLibrary(path)
            ... except OSError:
            ...     raise  SymupyLoadLibraryError("Library not found",path)

"""
# ============================================================================
# STANDARD  IMPORTS
# ============================================================================

import warnings

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================


class SymupyError(Exception):
    """General error exception, it can be raised at whatever moment,
        the class its intends to create a base class that intends to
        tag error situations dected by this module.

        :return: General ``SymupyError`` exception handler
        :rtype: SymupyError
    """

    def __init__(self, error_message: str = "", *args) -> None:
        super().__init__(error_message)
        self._extraargs = args if args else ("",)

    @property
    def extra_args(self):
        """suppelmentary arguments"""
        return self._extraargs

    @property
    def get_message(self):
        """message info"""
        mess = self.args if self.args else ("",)
        return mess

    def __str__(self) -> str:
        return "{0} ".format(*self.get_message) + "with args: {0}".format(
            *self.extra_args
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.get_message},{self.extra_args})"
        )


class SymupyWarning:
    """
       General warning exception, it can be raised at whatever moment, the
       class its intends to create a base class that intends to tag
       situations dected by this module.

        :param Exception: General ``SymupyWarning`` exception handler
        :type Exception: SymupyWarning
    """

    def __init__(self, warning_message: str) -> None:
        warnings.warn(warning_message)


class SymupyFileLoadError(SymupyError):
    """
        File Load Error exception handler, created for handling file situations
        in particular scenario situations

        :return: File Load Error
        :rtype: SymupyFileLoadError
    """

    def __init__(self, error_message: str, target_dir: str = "") -> None:
        super().__init__(error_message, target_dir)
        self._target_dir = target_dir

    def __str__(self) -> str:
        return "{0}".format(*self.get_message) + f" at: {self._target_dir}"

    def __repr__(self) -> str:
        return f"SymupyFileLoadError({self.get_message},{self._target_dir})"


class SymupyLoadLibraryError(SymupyError):
    """
        Load Library Error exception handler, created for handling library
        situations in particular when it is not possible to link the library
        and python

        :return: Library Load Error
        :rtype: SymupyLoadLibraryError
    """

    def __init__(self, error_message: str, target_dir: str = "") -> None:
        super().__init__(error_message, target_dir)
        self._target_dir = target_dir

    def __str__(self) -> str:
        return "{0}".format(*self.get_message) + f" at: {self._target_dir}"

    def __repr__(self) -> str:
        return f"SymupyLoadLibraryError({self.get_message},{self._target_dir})"


class SymupyVehicleCreationError(SymupyError):
    """
        Vehicle Creation Error exception handler, created for handling
        situations at runtime when the vehicles cannot be created within the
        network

        :return: Vehicle Creation Error
        :rtype: SymupyVehicleCreationError
    """

    def __init__(self, error_message: str, target_dir: str = "") -> None:
        super().__init__(error_message, target_dir)
        self._target_dir = target_dir

    def __str__(self) -> str:
        return "{0}".format(*self.get_message) + f" at: {self._target_dir}"

    def __repr__(self) -> str:
        return (
            f"SymupyVehicleCreationError({self.get_message},{self._target_dir})"
        )


class SymupyDriveVehicleError(SymupyError):
    """
        Vehicle Drive Error exception handler, created for handling situations
        at runtime when the vehicles cannot be created within the network

        :return: Vehicle Creation Error
        :rtype: SymupyDriveVehicleError
    """

    def __init__(self, error_message: str, target_dir: str = "") -> None:
        super().__init__(error_message, target_dir)
        self._target_dir = target_dir

    def __str__(self) -> str:
        return "{0}".format(*self.get_message) + f" at: {self._target_dir}"

    def __repr__(self) -> str:
        return f"SymupyDriveVehicleError({self.get_message},{self._target_dir})"



from setuptools import setup, find_packages

CLASSIFIERS = [
    "Development Status :: 1 - Planning",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Natural Language :: English",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: MacOS :: MacOS X",
    "Programming Language :: Python :: 3.7",
]

with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()

with open("HISTORY.md", "r") as history_file:
    HISTORY = history_file.read()

setup(
    name="symupy",
    version="0.4.2",
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        "numpy>=1.16",
        "lxml>=4.3.3",
        "xmltodict>=0.12",
        "networkx>=2.2",
        "matplotlib>=3.0.0",
        "sqlalchemy>=1.3.5",
        "pandas>=0.24",
        "scipy>=1.4.1",
        "click>=7.0",
        "python-decouple>=3.3",
    ],
    author="Andres Ladino",
    author_email="aladinoster@gmail.com",
    description="A module for Symuvia inside Python",
    url="https://github.com/symuvia/symupy",
    download_url="https://github.com/symuvia/symupy",
    include_package_data=True,
    long_description=LONG_DESCRIPTION + "\n\n" + HISTORY,
    long_description_content_type="text/markdown",
    keywords="traffic microsimulation",
    license="MIT",
    classifiers=CLASSIFIERS,
    zip_safe=False,
)



"""Expose specific functions for accessbility"""
from symupy.utils.tools import timer_func, logger_func, printer_time
from symupy.utils.parser import SimulatorRequest
from symupy.utils.configurator import Configurator



"""
    This module describes classes and objects to perform a runtime of a single scenario
"""

# ============================================================================
# STANDARD  IMPORTS
# ============================================================================

from itertools import chain
import click

# ============================================================================
# INTERNAL IMPORTS
# ============================================================================

from .states import Compliance, Connect, Initialize, PreRoutine, Query, Control, Push, PostRoutine, Terminate

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================

start_seq = ["compliance", "connect", "initialize"]
runtime_seq = ["preroutine", "query", "control", "push", "postroutine"]
end_seq = [
    "terminate",
]


class RuntimeDevice:
    """ 
        This class defines the runtime device describing a series of cyclic states required to be run 

        :return: Runtime Device for controlling states of the simulation runtime
        :rtype: RuntimeDevice
    """

    def __init__(self) -> None:
        click.echo("Runtime: Initialization")
        self.state = Compliance()  # Initial state

    def __logic(self, boolContinue: bool = True) -> dict:
        """ 
            Logic for state machine, transitions 

            * ``Compliance`` -> ``Connect``
            * ``Connect`` -> ``Initialize``
            * ``Initialize`` -> ``PreRoutine``
            * ``PreRoutine`` -> ``Query``
            * ``Query`` -> ``Control``
            * ``Control`` -> ``Push``
            * ``Push`` -> ``PostRoutine``
            * ``PostRoutine`` -> ``PreRoutine`` when ``boolContinue`` = True
            * ``PostRoutine`` -> ``Terminate`` when ``boolContinue`` = False
            * ``Terminate`` -> ``Terminate``

            :param boolContinue: Flag to determine continue looping, defaults to True
            :type boolContinue: bool, optional
            :return: State 
            :rtype: State
        """
        dct = {
            ("Compliance", True): self.state.on_event("Connect"),
            ("Connect", True): self.state.on_event("Initialize"),
            ("Initialize", True): self.state.on_event("PreRoutine"),
            ("PreRoutine", True): self.state.on_event("Query"),
            ("Query", True): self.state.on_event("Control"),
            ("Control", True): self.state.on_event("Push"),
            ("Push", True): self.state.on_event("PostRoutine"),
            ("PostRoutine", True): self.state.on_event("PreRoutine"),
            ("PostRoutine", False): self.state.on_event("Terminate"),
            ("Terminate", True): self.state.on_event("Terminate"),
            ("Terminate", False): self.state.on_event("Terminate"),
        }
        return dct.get((self.state.__class__.__name__, boolContinue), self.state)

    def reset_state(self) -> None:
        """
            Reset to initial state in case required 
        """
        self.state = Compliance()

    def next_state(self, cycle: bool = True) -> None:
        """
            Updates the state according to the state machine logic, possible states 

            * ``Compliance``
            * ``Connect``
            * ``Initialize``
            * ``PreRoutine``
            * ``Query``
            * ``Control``
            * ``Push``
            * ``PostRoutine``
            * ``Terminate``        

            :param cycle: Cycle parameter to return to PreRoutine state, defaults to True
            :type cycle: bool, optional
        """
        self.state = self.__logic(cycle)
